
structure CodGenC_back = struct

open CodGenC LR Say

infix  4 INIT
infix  3 --

(* code generate simple integer typedef *)

fun cg_typedef (name :string) (size:int) :codgen =
	let val celltype = tyname (uint_bysize size)
	    val _ = sayln VHigh (name ^ ":" ^ itoa size ^ " " ^ celltype)
	in TYPEDEF celltype name end handle e => intErr e 200209255 


(* model independent symbolic dictionary entry type *)

val struct_symbol = sg "struct struct_symbol { int id; const char * name; };\n\
		       \typedef struct struct_symbol symbol;\n"

(* code generate name macros for events (only for external) *)

fun cg_evmacros (eenv :eenv) (evtypes :evtypes) :codgen =
	let val lst      = SE.toList evtypes 
	    val ext      = List.filter(fn (ev,(i,ty)) => not i) lst
	    val names    = map #1 ext
	    fun def name = DEFINE (Symbol.name name) 
				(sg (itoa (SE.lookup eenv name)))
	in foldl (fn (na,e) => e -- def na) eoc names end

(* code generate symbolic names of events for debugging purposes *)
(* would be more efficient to do it in cg_evmacros, but this way *)
(* is cleaner.							 *)

fun cg_evdict debug (eenv :eenv) :codgen option =
	if not debug then NONE
	else let fun compare ((_,a),(_,b)) = Int.compare (a,b)
		 val lst  = Listsort.sort compare (SE.toList eenv)
		 val lst1 = map (quote o Symbol.name o #1) lst 
		 val inib = mk_commented_arrini lst1 1
		 val asize= itoa (length lst1)
	in SOME (ARRAY_DECL "const char *" "evdict" asize INIT inib) end

(* generate structure for single event type                  *)

fun cg_evtype (_,(_,[]))			= eoc
  | cg_evtype (name :symbol,et :evtype) :codgen =
	let val tynames = map tyname (#2 et)
	    fun g (tn,(i,e)) : int * string list = 
		(i+1, "\t" ^ tn ^ " f" ^ itoa i ^ ";\n" ::e)
	in 
		cg ["\nstruct Struct", Symbol.name name, " {\n"]
		-- cg (#2 (foldr g (0,[]) tynames))
		-- sg "};\n"
	end
(* generate structures for all event types *)

fun cg_evtypes (evtypes :evtypes) :codgen =
	let val lst = SE.toList evtypes
	in foldl (fn (et,e) => e -- cg_evtype et) eoc lst end

(* generate a union type of all events. Would more efficient *)
(* (but less clear) to do it in cg_evtypes.                  *)

fun cg_evunion (evtypes :evtypes) :codgen =
	let val lst = SE.toList evtypes
	    fun k ((n,(_,[])),e) = e
              | k ((n,(_,_)),(os,_)) = 
		(String.concat ["\t\tstruct Struct", Symbol.name n, 
			" _E_", Symbol.name n, ";\n"] ::os, true)
	    val (fields, nonempty) = foldl k ([],false) lst
	    val union = if nonempty then 
				sg "\tunion {\n" -- cg fields -- sg "\t} fields;\n"
			else eoc
	in sg "\nstruct StructEvent {\n\
			\\n\teventtag tag;\n" -- union -- sg "\n};\n\n"
		-- sg "\nextern struct StructEvent CurrEvent;\n"
	end

(* generate variable definition (and extern decl for external variable) *)

fun cg_vardecl (id:symbol) (static :bool) (ty :ty) (ini :value) 
							:codgen * codgen =
	let fun f id (SimpleType ty  ) = tyname ty ^ " " ^ Symbol.name id
	      | f id (ArrayType(ty,n)) = tyname ty ^ " " ^ Symbol.name id 
					     ^ "[" ^ itoa n ^ "]"
	    val dcl = f id ty
	    val def = dcl ^ " = " ^ dump_value ini 
	in if static then (cg ["static ",def, ";\n"], eoc)
	   else (cg [def,";\n"], cg ["extern ",dcl,";\n"])
	end

fun cg_vars  (venv :venv) (vartypes :vartypes) :codgen * codgen =
	let val tlst = SE.toList vartypes
	    fun f (v,e) = let val ini = SE.lookup venv (#1 v)
			      val internal :bool = #1 (#2 v)
			      val ty :ty = #2 (#2 v)
			      val (def,dcl) = cg_vardecl (#1 v) internal ty ini
			  in (#1 e -- def, #2 e --dcl ) end
	in foldl f (eoc,eoc) tlst end handle e => intErr e 200209251

(* generate constant definitions (preprocessor) *)

fun cg_consts (cenv :const env) :codgen =
	let val lst = map (fn (na,va) => (S.name na,dump_const va)) 
				(SE.toList cenv)
	    val cgs = map (fn (na,va) => DEFINE na (sg va)) lst 
	in foldr op-- eoc cgs end

(* code generate expression *)	

fun cg_exp (e :aexp) :codgen =
	case e of
		  Cst c => sg (dump_const c)
		| Access a => cg_acc a 
		| Prim1  (ope,e1) =>
			cg [dump_Prim1 ope, "("] -- cg_exp e1 -- sg ")"
		| Prim2  (ope,e1,e2) =>
			sg "(" -- cg_exp e1 -- sg ")"
				-- sg (dump_Prim2 ope) --
			sg "(" -- cg_exp e2 -- sg ")"
		| Call oe => cg_oexp oe
		| After (e1,e2) => 
			let val ce2 = cg_exp e2
			in cg_exp e1 -- sg "?" -- ce2 -- sg ":" -- ce2 end

(* code generate accesses - with substitution *)
		
and cg_acc (a :access) :codgen = 
	case a of AccessVar v => sg (Symbol.name v)
		| AccessArr (a,e) => 
			sg (Symbol.name a) -- sg "[" --  cg_exp e -- sg "]" 

(* generate output expressions - with substitution *)			
  
and cg_oexp ((f,[])  :oexp) :codgen = cg [Symbol.name f, "()"]
  | cg_oexp ((f,[p]) :oexp) :codgen = 
	cg [Symbol.name f, "("] -- cg_exp p -- sg ")"
  | cg_oexp ((f,h::t):oexp) :codgen =
	let val ch = cg_exp h
	    val ct = map cg_exp t
	in cg [Symbol.name f, "("] -- foldl (fn (f,s) => f -- sg "," -- s) 
		ch ct -- sg ")" 
	end

(* CG boolean expression *)

val cg_bexp = cg_exp

(* CG assignments *)

fun cg_assgn ((a,e) :assgn) :codgen =
	cg_acc a -- sg " = " -- cg_exp e 

(* code generate a guard. Guard should never be null as front end allocates   *)
(* "always" guard for all empty entries. Even this gurad is redundant in data *)
(* layouts with optional guards but in some layaouts (with packed guards) we  *)
(* need the value of "always"                                                 *)

fun cg_guard (g :guard) : codgen =
	let val ret = RETURN (cg_bexp 
			(foldl (fn (r,l) => Absyn.Prim2(Absyn.OpAnd,l,r)) 
			(hd g) (tl g)))
	in	
		sg "\t\t" -- ret -- sg "\n" 
		before assert (fn () => not (null g))
	end

(* code generate a guard dispatcher *)

fun cg_guards (markno :int) (guards :guards) : codgen option * codgen =
	if UList.length guards <= 1 then (* only always *)
		(NONE, UNDEF "SCOPE_GUARDS") 
	else let val lst    = UList.mapl cg_guard guards
		 val ids    = List.tabulate (length lst, fn x => CASE (itoa (x+markno)))
		 val cases1 = ListPair.map (op$) (ids, lst)
		 val cases2 = foldl (fn (x,e) => e -- sg "\t" -- x) eoc cases1
		 val body   = SWITCH "g" cases2 -- sg "\n\t\t return 1;"
		 val disp   = SOME (FUNC (sg "int eval ( const guardref g )") body)
		 val macro  = DEFINE "SCOPE_GUARDS" eoc
	in (disp, macro) end

(* find signals entry for given action *)


	(*  A HORRIBLE repetition, have to recompile signals to check the
	    address!!!  FIXME This is used in two places (here and
	    CF.cg_init).  Would be better to use references (ints) to signal
	    array already in LR, or make a scheduler compute it as additional
	    list of information. *)

fun signals2int (sigs :signals) ([]  :sexp list) :int = impossible "#200307230"
  | signals2int (sigs :signals) ([s] :sexp list) :int = impossible "#200307231"
  | signals2int (sigs :signals) (ss :sexp list) :int =
	let val css = map (LREventRef o #1) ss @ [LRNoEvent]
	in UList.find (css,sigs) end 
	handle UList.NotFound => impossible "#200209250"
	

(* code generate a signal action reference *)

fun cg_sigaction (eenv :eenv) (evnumber :int) (sigs :signals) 
					(signal  :sexp list) :codgen =
	case signal of 
		  []  => eoc 
		| [(s,_)] => sg "\t\tSET_RETURN_SIGNAL("
				-- ig (SE.lookup eenv s) -- sg ");\n"
		| ss  => let val vssignal = cg [itoa (signals2int sigs signal),
						" + EVENTNO + 1"]
			 in sg "\t\tSET_RETURN_SIGNAL(" 
				-- vssignal -- sg ");\n" 
			 end
	
(* code generate actions *)

fun cg_cmd (Expr  e) = cg_exp e
  | cg_cmd (Assgn a) = cg_assgn a

fun cg_cmds cmds :codgen =
	foldl (fn(a,e)=> e -- sg "\t\t" -- cg_cmd a -- sg ";\n") eoc cmds

fun cg_action (eenv :eenv) (evnumber :int) (sigs :signals)	
				(cmds,signals) :codgen = 
	let val ccmds  = cg_cmds cmds 
	    val vssignal = cg_sigaction eenv evnumber sigs signals
	in  ccmds -- vssignal  end

(* Code generate action dispatcher *)

fun cg_actions (markno :int) (eenv :eenv) (evnumber :int) (sigs :signals) 
					(actions :actions) :codgen =
	let val lst1   = rev (UList.mapl (cg_action eenv evnumber sigs) actions)
	    val lst2   = map (fn a=> a -- sg "\t\tbreak;\n") lst1
	    val ids    = List.tabulate (length lst2,fn x => CASE (itoa (x+markno)))
	    val cases1 = ListPair.map (op$) (ids, lst2)
	    val cases2 = foldl (fn (x,e) => e -- sg "\t" -- x) eoc cases1
	    val cases3 = List.tabulate(markno, 
				fn i => sg "\t" -- CASE (itoa i) eoc)
	    val cases4 = CodGen.concat cases3 -- cases2
	in
		sg "void exec ( const actionref a ) { \n\				
		   \\t#ifdef SCOPE_SIGNALQUEUE\n\
			\\tsignalref _vssignal = NOEVENT;\n\
			\\textern void enqueue1 ( eventtag );\n\
			\#ifdef SCOPE_SLIST\n\
			\\textern void enqueueList ( signalref );\n\
			\#endif\n\
		   \\t#endif\n\n\t" --
		SWITCH "a" cases4 --
		sg "\n\t#ifdef SCOPE_SIGNALQUEUE\n\
			\\t#ifdef SCOPE_SLIST\n\
			\\tif (_vssignal < NOEVENT) enqueue1(_vssignal);\n\
			\\telse if (_vssignal > NOEVENT)\n\
			\\t\tenqueueList (_vssignal-NOEVENT-1);\n\
			\\t#else\n\
			\\tif (_vssignal != NOEVENT) enqueue1(_vssignal);\n\
			\\t#endif\n\
		   \\t#endif\n\
		   \}\n"
	end

fun cg_externhdr (sysname :string) (externhdr :string option) :codgen =
	let val hdr = case externhdr of NONE => sysname ^ "_extern.h" 
				      | SOME s => s
	in INCLUDE hdr end

(* add named LR mnemonic to compund (encoded) value *)

fun compound_mnemonic (name :string) (ints :int list) :string list =
	if null ints then []
	else String.concat [name, "(", itoa (length ints), ") ", itoa (hd ints)]
		::mitoa (tl ints)


(* compile a LR instructions/fields *)

fun comp_marker STMRK   = "STMRK"
  | comp_marker HMRK    = "HMRK"
  | comp_marker IMRK    = "IMRK" 
  | comp_marker FPSCOP  = "FPSCOP"
  | comp_marker NFPSCOP = "NFPSCOP"

(* pack positive and negative counters *)

fun pack_pcnc (sizes :lrsizes) (pc :int, nc :int) :string list =
	let val ints = pack2 (pc,nc) (1,1) (#andstref sizes)
	in compound_mnemonic "PCNC" ints end

fun cg_unmangle_pcnc (sizes :lrsizes) :codgen =
	DEFINE "UNMANGLE_PCNC(ptr,i,j)" 
		(access2 (1,1) (#andstref sizes))

(* compile an array in LR *)

fun cg_arraylr (name :string) (tyname :string) (arr :lratom list) (comp_lr) 
									:codgen =
	let val lst      = foldr comp_lr [] arr
	    val iniblock = mk_commented_arrini lst 5 
	    val asize    = itoa (length lst)
	in ARRAY_DECL tyname name asize INIT iniblock end
	handle e => intErr e 200209256

(* compile signal groups. return signals array (option) and #define *)
(* SCOPE_SLIST for                                                  *)

fun cg_signals (signals :signals) (comp_lr) :codgen option * codgen =
	if UList.null signals then (NONE, UNDEF "SCOPE_SLIST")
	else let val arr = List.concat (UList.toList signals)
		 val def = cg_arraylr "signals" "const eventtag" arr comp_lr
	     in (SOME def, DEFINE "SCOPE_SLIST" eoc) end

(* generate configuration vector/set *)

fun cg_conf (name :string) (width :int) :codgen =
	let val iniblock = mk_commented_arrini (List.tabulate 
				(width, fn _ => "STMRK")) 10
	in VAR_DECL "static conf" name INIT iniblock end 

(* code generate symbolic dictionaries if needed. *)
(* These are used for debugging purposes.         *)

fun cg_intenv (dname :string) (dict :int env) :codgen =
	let fun mkstruct (a,b) = 
			String.concat ["{",itoa b,",",quote (S.name a),"}"]
		 val null = mkstruct (S.symbol "ERROR",0)
		 val lst  = foldr (fn (x,e) => mkstruct x :: e) 
				[null] (SE.toList dict) 
		 val inib = mk_commented_arrini lst 1
		 val asize = itoa (length lst)
	in ARRAY_DECL "const symbol" dname asize INIT inib end

(* transition index *)

fun cg_tranidx tranidx eenv (sizes :lrsizes) (tranlen :int) :codgen * codgen =
	let fun mkoffsets (ev,i) = (SE.lookup eenv ev,i)
	    fun compare ((a,_),(b,_)) = Int.compare (a,b)
	    val lst1 = (#evnumber sizes, tranlen)
			:: map mkoffsets (SE.toList tranidx)
	    val lst2 = map #2 (Listsort.sort compare lst1)
	    val iniblock = mk_commented_arrini (mitoa lst2) 10
	    val asize = itoa (length lst2)
	in ( ARRAY_DECL "const tranref" "tranidx" asize INIT iniblock, 
	     cg_typedef "tranref" (mintysize (List.last lst2)))
	end handle e => intErr e 200307221
	    (* the last should be max... YUCK: this is slow *)



(* Code generate signal queue #defines and array if needed *)

fun cg_queue (size :int option) :codgen =
	case size of 
		  NONE   => UNDEF "SCOPE_SIGNALQUEUE" 
				(* SET_RETURN_SIGNAL should never be needed *)
				(* in this case                             *)
		| SOME n => 
			let val lst      = List.tabulate (n,fn _ =>"NOEVENT")
			    val iniblock = mk_commented_arrini lst 7
			in 
				DEFINE "SCOPE_SIGNALQUEUE" eoc
				-- DEFINE "QUEUE_LENGTH" (ig n)
				-- ARRAY_DECL "static eventtag" "queue" "QUEUE_LENGTH"
					INIT iniblock 
				-- DEFINE "SET_RETURN_SIGNAL(s)" 
					(sg "{_vssignal = (s);}")
			end

(* generate transition table. I cannot use cg_arraylr because I need to know *)
(* the size of trans array for type of traniter. Retvalues are: the array    *)
(* code and number of cells in the array                                     *)
(* This is reused across code generators as comp_lr is supllied externally   *)

fun cg_trans comp_lr (tenv :tenv) :codgen * int =
	let val lst  = List.concat (map #2 (SE.toList tenv))
	    val ilst = foldr comp_lr [] lst
	    val iniblock = mk_commented_arrini ilst 5
	    val asize = itoa (length ilst)
	in (DEFINE "TRANS_MAX" (sg asize) -- eol -- eol
		-- ARRAY_DECL "const transcell" "trans" "TRANS_MAX" 
			INIT iniblock,
	    length ilst)
	end handle e => intErr e 200209252 



end
